# SPDX-License-Identifier: MIT

from __future__ import print_function

import argparse
import logging
import os
import shutil
import sys
import traceback
import typing
import warnings

from typing import Optional, TextIO, Union


# isort:maintain_block
if False:  # TYPE_CHECKING
    from typing import Type  # not in 3.5.{0,1}
# isort:end_maintain_block


from . import IncompleteInstallationWarning, InstallException, build, install  # isort:skip # until isort 5.0.0


logger = logging.getLogger('install.main')


def _showwarning(message, category, filename, lineno, file=None, line=None):
    # type: (Union[Warning, str], Type[Warning], str, int, Optional[TextIO], Optional[str]) -> None
    prefix = 'WARNING'
    where = '{}:{}'.format(filename, lineno)
    name = '({})'.format(category.__name__)
    if sys.stdout.isatty():
        prefix = '\33[93m' + prefix + '\33[0m'
        where = '\33[2m' + where + '\33[0m'
        name = '\33[33m\33[7m' + name + '\33[0m'
    if not file:
        if sys.version_info >= (3,):
            file = sys.stderr
        else:
            file = typing.cast(TextIO, sys.stderr)
    print('{} {} {} {}'.format(prefix, where, name, message), file=file)


warnings.showwarning = _showwarning


def _error(msg, code=1):  # type: (str, int) -> None
    prefix = 'ERROR'
    if sys.stdout.isatty():
        prefix = '\33[91m' + prefix + '\33[0m'
    print('{} {}'.format(prefix, msg))
    exit(code)


if __name__ == '__main__':  # noqa: C901
    sys.argv[0] = 'python -m install'
    parser = argparse.ArgumentParser()
    parser.add_argument('wheel', nargs='?',
                        type=str,
                        help='wheel file to install')
    parser.add_argument('--verbose', '-v',
                        action='store_true',
                        help='enable verbose output')
    parser.add_argument('--optimize', '-o', nargs='*', metavar='level',
                        type=int, default=[0, 1, 2],
                        help='optimization level(s) (default=0, 1, 2)')
    parser.add_argument('--destdir', '-d', metavar='/',
                        type=str, default='/',
                        help='destination directory')
    parser.add_argument('--verify-dependencies', '-t',
                        action='store_true',
                        help='check if the dependencies are met')
    # build/install separation
    parser.add_argument('--cache', '-c',
                        action='store_true',
                        help='generate the installation cache')
    parser.add_argument('--skip-build', '-s',
                        action='store_true',
                        help='skip the cache building step, requires cache to be present already')
    # warnings
    parser.add_argument('--ignore-incomplete-installation-warnings', '-w',
                        action='store_true',
                        help='stop treating incomplete installation warnings as errors')
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)

    if not args.ignore_incomplete_installation_warnings:
        warnings.simplefilter('error', IncompleteInstallationWarning)

    cache_dir = '.install-cache'
    pkg_cache_dir = os.path.join(cache_dir, 'pkg')
    entrypoints_cache_dir = os.path.join(cache_dir, 'entrypoints')

    if args.cache and args.skip_build:
        _error("--cache and --skip-build can't be used together, choose one")

    if not args.wheel and not args.skip_build:
        _error('Missing argument: wheel')

    if os.path.exists(cache_dir):
        if os.path.isdir(cache_dir):
            if not args.skip_build:
                logger.debug('Cache directory exists, removing')
                shutil.rmtree(cache_dir)
        else:
            _error("The cache path ('{}') exists and it's not a directory".format(cache_dir))

    # Build cache
    if not args.skip_build:
        try:
            build(args.wheel, cache_dir, args.optimize, args.verify_dependencies)
        except InstallException as e:
            _error(str(e))
        except Exception as e:
            print(traceback.format_exc())
            _error(str(e))

    # Install to destination
    if not args.cache:
        if not os.path.isdir(cache_dir):
            _error('Missing installation cache (hint: python -m install --cache [ ... ])')

        try:
            install(cache_dir, args.destdir)
        except InstallException as e:
            _error(str(e))
        except Exception as e:
            print(traceback.format_exc())
            _error(str(e))



"""Parsing and handling :pep:`376` RECORD files."""

import base64
import csv
import hashlib

from installer._compat.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Iterator, Optional
    from installer._compat.typing import FSPath


__all__ = [
    "Hash",
    "Record",
    "InvalidRecord",
    "parse_record_file",
]


class InvalidRecord(Exception):
    """Raised when a Record is not valid, due to improper element values or count."""

    def __init__(self, elements, issues):  # noqa: D107
        super(InvalidRecord, self).__init__(", ".join(issues))
        self.issues = issues
        self.elements = elements

    def __repr__(self):
        return "InvalidRecord(elements={!r}, issues={!r})".format(
            self.elements, self.issues
        )


class Hash(object):
    """Represents the "hash" element of a Record."""

    def __init__(self, name, value):
        # type: (str, str) -> None
        """Construct a ``Hash`` object.

        Most consumers should use :py:meth:`Hash.parse` instead, since no
        validation or parsing is performed by this constructor.

        :param name: name of the hash function
        :param value: hashed value
        """
        self.name = name
        self.value = value

    def __repr__(self):
        # type: () -> str
        return "Hash(name={!r}, value={!r})".format(self.name, self.value)

    def validate(self, data):
        # type: (bytes) -> bool
        """Validate that ``data`` matches this instance.

        :param data: Contents of the file.
        :return: Whether ``data`` matches the hashed value.
        """
        digest = hashlib.new(self.name, data).digest()
        value = base64.urlsafe_b64encode(digest).decode("ascii").rstrip("=")
        return self.value == value

    @classmethod
    def parse(cls, h):
        # type: (str) -> Hash
        """Build a Hash object, from a "name=value" string.

        This accepts a string of the format for the second element in a record,
        as described in :pep:`376`.

        Typical usage::

            Hash.parse("sha256=Y0sCextp4SQtQNU-MSs7SsdxD1W-gfKJtUlEbvZ3i-4")

        :param h: a name=value string
        """
        name, value = h.split("=", 1)
        return cls(name, value)


class Record(object):
    """Represents a single record in a RECORD file.

    A list of :py:class:`Record` objects fully represents a RECORD file.
    """

    def __init__(self, path, hash_, size):
        # type: (FSPath, Optional[Hash], Optional[int]) -> None
        r"""Construct a ``Record`` object.

        Most consumers should use :py:meth:`Record.from_elements`, since no
        validation or parsing is performed by this constructor.

        :param path: file's path
        :param hash\_: hash of the file's contents
        :param size: file's size in bytes
        """
        super(Record, self).__init__()

        self.path = path
        self.hash_ = hash_
        self.size = size

    def __repr__(self):
        # type: () -> str
        return "Record(path={!r}, hash_={!r}, size={!r})".format(
            self.path, self.hash_, self.size,
        )

    def validate(self, data):
        # type: (bytes) -> bool
        """Validate that ``data`` matches this instance.

        :param data: Contents of the file corresponding to this instance.
        :return: whether ``data`` matches hash and size.
        """
        if self.size is not None and len(data) != self.size:
            return False

        if self.hash_:
            return self.hash_.validate(data)

        return True

    @classmethod
    def from_elements(cls, path, hash_, size):
        # type: (FSPath, str, str) -> Record
        r"""Build a Record object, from values of the elements.

        Typical usage::

            reader = csv.reader(f)
            for row in reader:
                record = Record.from_elements(row[0], row[1], row[2])

        Meaning of each element is specified in :pep:`376`.

        :param path: first element (file's path)
        :param hash\_: second element (hash of the file's contents)
        :param size: third element (file's size in bytes)
        :raises InvalidRecord: if any element is invalid
        """
        # Validate the passed values.
        issues = []

        if not path:
            issues.append("`path` cannot be empty")

        if hash_:
            try:
                hash_value = Hash.parse(hash_)  # type: Optional[Hash]
            except ValueError:
                issues.append("`hash` does not follow the required format")
        else:
            hash_value = None

        if size:
            try:
                size_value = int(size)  # type: Optional[int]
            except ValueError:
                issues.append("`size` cannot be non-integer")
        else:
            size_value = None

        if issues:
            raise InvalidRecord(elements=(path, hash_, size), issues=issues)

        return cls(path=path, hash_=hash_value, size=size_value,)


def parse_record_file(rows):
    # type: (Iterator[str]) -> Iterator[Record]
    """Parse a :pep:`376` RECORD.

    :param rows: iterator providing lines of a RECORD.
    """
    reader = csv.reader(rows, delimiter=",", quotechar='"', lineterminator="\n")
    for row_index, elements in enumerate(reader):
        if len(elements) != 3:
            message = "Row Index {}: expected 3 elements, got {}".format(
                row_index, len(elements)
            )
            raise InvalidRecord(elements=elements, issues=[message])

        record = Record.from_elements(elements[0], elements[1], elements[2])
        yield record



# SPDX-License-Identifier: MIT

from __future__ import print_function

import compileall
import configparser
import fileinput
import logging
import os
import pickle
import platform
import py_compile
import re
import shutil
import sys
import sysconfig
import warnings
import zipfile

import install._vendor as _vendor  # noqa: F401


'''
python-install - A simple, correct PEP427 wheel installer
'''
__version__ = '0.0.3'


if sys.version_info >= (3, 5) or (sys.version_info < (3,) and sys.version_info >= (2, 7)):
    from typing import Any, Dict, List

if sys.version_info < (3,):
    FileNotFoundError = IOError
    PermissionError = OSError
    FileExistsError = OSError


_SUPPORTED_WHEEL_VERSION = (1, 0)

_WHEEL_NAME_REGEX = re.compile(r'(?P<distribution>.+)-(?P<version>.+)'
                               r'(-(?P<build_tag>.+))?-(?P<python_tag>.+)'
                               r'-(?P<abi_tag>.+)-(?P<platform_tag>.+).whl')

logger = logging.getLogger('install')


class IncompleteInstallationWarning(RuntimeWarning):
    '''
    Some installation step was not handled, potentially resulting in an
    incomplete instalation
    '''


class InstallWarning(RuntimeWarning):
    pass


class InstallException(Exception):
    pass


def _destdir_path(destdir, lib):  # type: (str, str) -> str
    path = sysconfig.get_path(lib)
    if not path:
        raise InstallException("Couldn't find {}".format(lib))
    return os.path.join(destdir, os.sep.join(path.split(os.sep)[1:]))


def _read_wheel_metadata(dist_info_path):  # type: (str) -> Dict[str, str]
    metadata = {}
    with open(os.path.join(dist_info_path, 'WHEEL')) as f:
        for line in f:
            entry = line.split(':')
            if len(entry) == 2:  # throw error?
                metadata[entry[0].strip()] = entry[1].strip()
    return metadata


if sys.version_info >= (3, 8):

    def _copy_dir(src, dst, ignore=[]):  # type: (str, str, List[str]) -> None
        shutil.copytree(src, dst, dirs_exist_ok=True, ignore=lambda *_: ignore)

else:

    def _copy_dir(src, dst, ignore=[]):  # type: (str, str, List[str]) -> None
        from distutils.dir_util import copy_tree
        for node in os.listdir(src):
            if node in ignore:
                continue
            path = os.path.join(src, node)
            root = os.path.join(dst, node)
            if os.path.isdir(path):
                copy_tree(path, root)
            else:
                shutil.copy2(path, root)


def _generate_entrypoint_scripts(file, dir):  # type: (str, str) -> None
    entrypoints = configparser.ConfigParser()
    entrypoints.read(file)
    if 'console_scripts' in entrypoints:
        if not os.path.exists(dir):
            os.mkdir(dir)

        import installer.scripts

        for name, backend in entrypoints['console_scripts'].items():
            package, call = backend.split(':')

            script = installer.scripts.Script(name, package, call, section='console')
            name, data = script.generate(sys.executable, kind='posix')

            with open(os.path.join(dir, name), 'wb') as f:
                f.write(data)


def _replace_shebang(dir, interpreter):  # type: (str, str) -> None
    scripts = [os.path.join(dir, script) for script in os.listdir(dir)]

    for script in scripts:
        if not os.path.isfile(script):
            raise InstallException('Script is not a file: {}'.format(script))

    # Python 2 does not support fileinput as a contex manager
    f = fileinput.input(scripts, inplace=True)
    for line in f:
        if f.isfirstline():
            line = re.sub(r'^#!python', '#!{}'.format(interpreter), line)
        print(line, end='')
    f.close()


def _check_requirement(requirement_string):  # type: (str) -> bool
    import packaging.requirements

    if sys.version_info >= (3, 8):
        from importlib import metadata as importlib_metadata
    else:
        import importlib_metadata

    req = packaging.requirements.Requirement(requirement_string)

    if req.marker and not req.marker.evaluate():
        return True

    try:
        version = importlib_metadata.version(req.name)
        metadata = importlib_metadata.metadata(req.name)
    except importlib_metadata.PackageNotFoundError:
        return False

    metadata_extras = metadata.get_all('Provides-Extra') or []
    for extra in req.extras:
        if extra not in metadata_extras:
            return False

    if req.specifier:
        return req.specifier.contains(version)

    return True


def _verify_compability(dir, verify_dependencies=False):  # type: (str, bool) -> None
    try:
        import packaging.specifiers

        if sys.version_info >= (3, 8):
            from importlib import metadata as importlib_metadata
        else:
            import importlib_metadata

        dist = importlib_metadata.Distribution.at(dir)

        py_ver = dist.metadata['Requires-Python']
        if py_ver:
            py_vers = py_ver.split(',')
            for ver in py_vers:
                py_spec = packaging.specifiers.Specifier(ver)
                if platform.python_version() not in py_spec:
                    raise InstallException('Incompatible python version, needed: {}'.format(py_ver))

        if verify_dependencies:
            for req in dist.metadata.get_all('Requires-Dist') or []:
                if not _check_requirement(req):
                    raise InstallException('Missing dependency: {}'.format(req))

    except ImportError as e:
        warnings.warn('{}: Platform/Python tags were not verified for compatibility'.format(e), InstallWarning)


def _save_pickle(dir, name, data):  # type: (str, str, Any) -> None
    with open(os.path.join(dir, name + '.pickle'), 'wb') as f:
        pickle.dump(data, f)


def _load_pickle(dir, name):  # type: (str, str) -> Any
    with open(os.path.join(dir, name + '.pickle'), 'rb') as f:
        return pickle.load(f)


def parse_name(name):  # type: (str) -> Dict[str, str]
    match = _WHEEL_NAME_REGEX.match(name)
    if not match:
        raise InstallException('Invalid wheel name: {}'.format(name))
    return match.groupdict()


def build(wheel, cache_dir, optimize=[0, 1, 2], verify_dependencies=False):  # type: (str, str, List[int], bool) -> None
    pkg_cache_dir = os.path.join(cache_dir, 'pkg')
    scripts_cache_dir = os.path.join(cache_dir, 'scripts')
    wheel_info = parse_name(os.path.basename(wheel))
    package = '{}-{}'.format(wheel_info['distribution'], wheel_info['version'])
    dist_info = os.path.join(pkg_cache_dir, '{}.dist-info'.format(package))
    entrypoints_file = os.path.join(dist_info, 'entry_points.txt')
    scripts_dir = os.path.join(pkg_cache_dir, '{}.data'.format(package), 'scripts')

    with zipfile.ZipFile(wheel) as wheel_zip:
        wheel_zip.extractall(pkg_cache_dir)

    metadata = _read_wheel_metadata(dist_info)

    if tuple(map(int, metadata['Wheel-Version'].split('.'))) > _SUPPORTED_WHEEL_VERSION:
        raise InstallException('Unsupported wheel version: {}'.format(metadata['Wheel-Version']))

    _verify_compability(dist_info, verify_dependencies)

    if sys.version_info >= (3,):
        for level in optimize:
            logger.debug('Optimizing for {}'.format(level))
            if sys.version_info >= (3, 7):
                compileall.compile_dir(pkg_cache_dir, optimize=level,
                                       invalidation_mode=py_compile.PycInvalidationMode.CHECKED_HASH)
            else:
                compileall.compile_dir(pkg_cache_dir, optimize=level)
    elif optimize:
        compileall.compile_dir(pkg_cache_dir)

    if os.path.isfile(entrypoints_file):
        _generate_entrypoint_scripts(entrypoints_file, scripts_cache_dir)

    if os.path.isdir(scripts_dir):
        _replace_shebang(scripts_dir, sys.executable)

    _save_pickle(cache_dir, 'wheel-info', wheel_info)
    _save_pickle(cache_dir, 'metadata', metadata)

    # TODO: verify checksums


def install(cache_dir, destdir):  # type: (str, str) -> None
    def destdir_path(lib):  # type: (str) -> str
        return _destdir_path(destdir, lib)

    wheel_info = _load_pickle(cache_dir, 'wheel-info')
    metadata = _load_pickle(cache_dir, 'metadata')

    pkg_cache_dir = os.path.join(cache_dir, 'pkg')
    scripts_cache_dir = os.path.join(cache_dir, 'scripts')
    pkg_data_dir_name = '{}-{}.data'.format(wheel_info['distribution'], wheel_info['version'])
    pkg_data_dir = os.path.join(cache_dir, pkg_data_dir_name)

    pkg_dir = destdir_path('purelib' if metadata['Root-Is-Purelib'] == 'true' else 'platlib')

    _copy_dir(pkg_cache_dir, pkg_dir, ignore=['purelib', 'platlib', pkg_data_dir_name])
    for lib in ['purelib', 'platlib']:
        target = os.path.join(pkg_cache_dir, lib)
        if os.path.isdir(target):
            _copy_dir(target, destdir_path(lib))
    if os.path.isdir(pkg_data_dir):
        for node in os.listdir(pkg_data_dir):
            target = os.path.join(pkg_cache_dir, node)
            if node == 'purelib':
                _copy_dir(target, destdir_path('purelib'))
            if node == 'platlib':
                _copy_dir(target, destdir_path('platlib'))
            if node == 'scripts':
                _copy_dir(target, destdir_path('scripts'))
            # TODO: headers, data -- is this a direct mapping to sysconfig? does it need specific path handling?
            else:
                warnings.warn('Unhandled data folder: {}'.format(node), IncompleteInstallationWarning)

    if os.path.isdir(scripts_cache_dir):
        _copy_dir(scripts_cache_dir, destdir_path('scripts'))
    # TODO: update dist-info/RECORD



"""Generate executable scripts, on various platforms."""

import io
import shlex
import zipfile

from installer import _scripts
from installer._compat import importlib_resources
from installer._compat.typing import TYPE_CHECKING, Binary, Text

if TYPE_CHECKING:
    from typing import Literal, Mapping, Optional, Tuple

    LauncherKind = Literal["posix", "win-ia32", "win-amd64", "win-arm"]
    ScriptSection = Literal["console", "gui"]


__all__ = ["InvalidScript", "Script"]


_ALLOWED_LAUNCHERS = {
    ("console", "win-ia32"): "t32.exe",
    ("console", "win-amd64"): "t64.exe",
    ("console", "win-arm"): "t_arm.exe",
    ("gui", "win-ia32"): "w32.exe",
    ("gui", "win-amd64"): "w64.exe",
    ("gui", "win-arm"): "w_arm.exe",
}  # type: Mapping[Tuple[ScriptSection, LauncherKind], str]

_SCRIPT_TEMPLATE = """\
# -*- coding: utf-8 -*-
import re
import sys
from {module} import {import_name}
if __name__ == "__main__":
    sys.argv[0] = re.sub(r"(-script\\.pyw|\\.exe)?$", "", sys.argv[0])
    sys.exit({func_path}())
"""


def _is_executable_simple(executable):
    # type: (Binary) -> bool
    if b" " in executable:
        return False
    shebang_length = len(executable) + 3  # Prefix #! and newline after.
    # According to distlib, Darwin can handle up to 512 characters. But I want
    # to avoid platform sniffing to make this as platform agnostic as possible.
    # The "complex" script isn't that bad anyway.
    return shebang_length <= 127


def _quote_compat(s):  # pragma: no cover
    # type: (Text) -> Text
    """Fallback implementation for ``shlex.quote`` (for Python 2)."""
    return u"'" + s.replace(u"'", u"'\"'\"'") + u"'"


def _build_shebang(executable, forlauncher):
    # type: (Text, bool) -> Binary
    """Build a shebang line.

    The non-launcher cases are taken directly from distlib's implementation,
    which tries its best to account for command length, spaces in path, etc.

    https://bitbucket.org/pypa/distlib/src/58cd5c6/distlib/scripts.py#lines-124
    """
    executable_bytes = executable.encode("utf-8")
    if forlauncher:  # The launcher can just use the command as-is.
        return b"#!" + executable_bytes
    if _is_executable_simple(executable_bytes):
        return b"#!" + executable_bytes

    # Shebang support for an executable with a space in it is under-specified
    # and platform-dependent, so we use a clever hack to generate a script to
    # run in ``/bin/sh`` that should work on all reasonably modern platforms.
    # Read the following message to understand how the hack works:
    # https://github.com/pradyunsg/installer/pull/4#issuecomment-623668717

    quote = getattr(shlex, "quote", _quote_compat)
    quoted = quote(executable).encode("utf-8")
    # I don't understand a lick what this is trying to do.
    return b"#!/bin/sh\n'''exec' " + quoted + b' "$0" "$@"\n' + b"' '''"


class InvalidScript(ValueError):
    """Raised if the user provides incorrect script section or kind."""


class Script(object):
    """Describes a script based on an entry point declaration."""

    __slots__ = ("name", "module", "attr", "section")

    def __init__(self, name, module, attr, section):
        # type: (str, str, str, ScriptSection) -> None
        """Construct a Script object.

        :param name: name of the script
        :param module: module path, to load the entry point from
        :param attr: final attribute access, for the entry point
        :param section: Denotes the "entry point section" where this was specified.
            Valid values are ``"gui"`` and ``"console"``.
        :type section: str

        """
        self.name = name
        self.module = module
        self.attr = attr
        self.section = section

    def __repr__(self):
        # type: () -> str
        return "Script(name={!r}, module={!r}, attr={!r}".format(
            self.name, self.module, self.attr,
        )

    def _get_launcher_data(self, kind):
        # type: (LauncherKind) -> Optional[Binary]
        if kind == "posix":
            return None
        key = (self.section, kind)
        try:
            name = _ALLOWED_LAUNCHERS[key]
        except KeyError:
            error = "{!r} not in {!r}".format(key, sorted(_ALLOWED_LAUNCHERS))
            raise InvalidScript(error)
        return importlib_resources.read_binary(_scripts, name)

    def generate(self, executable, kind):
        # type: (str, LauncherKind) -> Tuple[str, Binary]
        """Generate a launcher for this script.

        :param executable: Path to the executable to invoke.
        :param kind: Which launcher template should be used.
            Valid values are ``"posix"``, ``"win-ia32"``, ``"win-amd64"`` and
            ``"win-arm"``.
        :type kind: str

        :raises InvalidScript: if no appropriate template is available.
        :return: The name and contents of the launcher file.
        """
        launcher = self._get_launcher_data(kind)
        shebang = _build_shebang(executable, forlauncher=bool(launcher))
        code = _SCRIPT_TEMPLATE.format(
            module=self.module,
            import_name=self.attr.split(".")[0],
            func_path=self.attr,
        ).encode("utf-8")

        if launcher is None:
            return (self.name, shebang + b"\n" + code)

        stream = io.BytesIO()
        with zipfile.ZipFile(stream, "w") as zf:
            zf.writestr("__main__.py", code)
        name = "{}.exe".format(self.name)
        data = launcher + shebang + b"\n" + stream.getvalue()
        return (name, data)



"""Static Typing Support logic.

This is shim logic allows this library to use mypy, for static type analysis, while:

- not importing the 'typing' module at runtime.
- maintaining Python 2 support
- not completely confusing mypy :)

This module also provides `FSPath`, a type annotation for Path-like values, which
can only be imported inside an `if TYPE_CHECKING`, and used only in type comments.
"""

__all__ = ["Text", "Binary", "FSPath", "TYPE_CHECKING"]

import sys

try:  # pragma: no cover
    from typing import TYPE_CHECKING
except ImportError:  # pragma: no cover
    TYPE_CHECKING = False

if sys.version_info >= (3,):  # pragma: no cover
    Binary = bytes
    Text = str
else:  # pragma: no cover
    Binary = str
    Text = unicode  # noqa

# FSPath declaration (only for Python 2 support)
if TYPE_CHECKING:
    if sys.version_info[:2] >= (3, 4):
        from typing import Union
        from pathlib import Path

        FSPath = Union[Text, Path]
    else:
        FSPath = Text



"""Compatibility cover for importlib.resources, for older Python versions."""

from __future__ import absolute_import

import sys

__all__ = ["read_binary"]

if sys.version_info >= (3, 7):  # pragma: no cover
    from importlib.resources import read_binary
else:  # pragma: no cover
    from importlib_resources import read_binary
